from . import components
from ._extensions import set_extensions, get_extensions, get_doc_extensions, get_span_extensions, get_token_extensions
from ._version import __version__
from .util import load

set_extensions()
